#!/usr/bin/env python

"""
bookmarksapplet.py

Python panel applet to display the user's bookmarks in a menu.

Copyright (C) 2003-2006 David Boddie <david@boddie.org.uk>

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
"""

from qt import *
import qtxml

from kdecore import i18n, KAboutData, KGlobal, KURL
from kdeui import KAboutApplication, KMainWindow, KMessageBox, KPanelApplet, \
                  KPopupMenu
from kparts import KParts

import os


# Define an initialisation function similar to that used to initialise
# C++ panel applets.

def init(parent, configFile, name):

    """applet = init_applet(parent, configFile, name)
    
    An initialisation function which corresponds to the one used by C++
    applet authors.
    
    parent is the parent widget for the new applet widget.
    
    configFile is a string containing the name of the configuration file
    allocated to this instance of the applet.
    
    name is the name associated with this applet; for Python applets, it
    will be the name of the general launching library so you may wish to
    change it.
    """
    
    # Standard opening to this function.
    
    KGlobal.locale().insertCatalogue(name)
    
    # Create an instance of the applet.
    
    # Note that the KPanelApplet.About bit must be set for there to be
    # an About entry in the applet handle's menu.
    
    applet = AppletClass(
        configFile, KPanelApplet.Normal, KPanelApplet.About, parent,
        "bookmarksapplet"
        )
    
    return applet


# Define a class based on the KPanelApplet class.

class AppletClass(KPanelApplet):

    """AppletClass(KPanelApplet)
    
    applet = AppletClass(config_file, panel_type, panel_actions, parent, name)
    
    Create an applet instance with the given properties.
    
    panel_type is either KPanelApplet.Normal or KPanelApplet.Stretch
    
    panel_actions is a bitwise OR of
    
        KPanelApplet.About
        KPanelApplet.Help
        KPanelApplet.Preferences
        KPanelApplet.ReportBug
    
    name is the name of the applet.
    """
    
    def __init__(self, config_file, panel_type, panel_actions, parent, name):
    
        KPanelApplet.__init__(
            self, config_file, panel_type, panel_actions, parent, name
            )
        
        # Record the values of the arguments passed to this constructor
        # for possible future use.
        
        self.name = name
        self.parent = parent
        self.config_file = config_file
        
        # Define a variable to contain the KAboutData instance if/when it
        # is generated.
        
        self._about_data = None
        
        # Define applet information. Maybe I should define some translation
        # strings somewhere.
        
        self.version = "0.20 (30th August 2006)"
        self.program_name = str(i18n("Bookmarks"))
        self.program_description = str(
            i18n(
                "Show Konqueror bookmarks in a pop-up menu."
                )
            )
        
        self.thanks = str(i18n(
            "Thanks to Jim Bublitz and Phil Thompson for providing advice " + \
            "on using PyKDE and SIP."
            ))
        
        self.authors = \
        [
            ( "David Boddie", str(i18n("Applet author")),
              "david@boddie.org.uk", "http://www.boddie.org.uk/" )
        ]
        
        # Record the path to this module. We could obtain this from the general
        # launcher, since it has to locate this module anyway.
        
        self.module_path = __file__
        
        # Find a pixmap in the directory where this module resides.
        
        appdir, appmodule = os.path.split(self.module_path)
        
        pixmap_file = os.path.join(appdir, "bookmarks" + os.extsep + "png")
        
        if os.path.exists(pixmap_file):
        
            pixmap = QPixmap(pixmap_file)
            self.image = pixmap.convertToImage()
        
        else:
        
            self.setBackgroundMode(self.NoBackground)
            self.brush = QBrush(self.darkRed)
    
    def widthForHeight(self, height):
    
        """width = widthForHeight(self, height)
        
        For a given height, return an integer value for the required width
        of the applet.
        """
        
        # width = 2 * height/3
        return int(2 * height/3.0)
    
    def heightForWidth(self, width):
    
        """height = heightForWidth(self, width)
        
        For a given width, return an integer value for the required height
        of the applet.
        """
        
        # width = 2 * height/3
        # => height = 3 * width/2
        return int(3 * width/2.0)
    
    def about(self):
    
        """about(self)
        
        Called when the About entry is activated from the menu opened
        from the tiny arrow button on the applet's handle.
        
        The appropriate flag must have been set when this class was
        instantiated.
        """
        
        if self._about_data is None:
        
            # The QStrings returned from the i18n function need to be
            # converted to strings for the following constructor.
            # The program name and program description strings must also be
            # translated and defined elsewhere rather than being translated
            # and converted to strings inline.
            
            self._about_data = KAboutData(
                self.name, self.program_name, self.version,
                self.program_description,
                KAboutData.License_GPL, "(c) 2003-2006, David Boddie",
                self.thanks
                )
            
            # You _must_ set the license text if the License_Custom
            # value is used in the constructor.
            #self._about_data.setLicenseText(self.license)
            
            # It looks like all arguments are required for the addAuthor
            # method of the KAboutData class.
            
            for author, role, email, website in self.authors:
            
                self._about_data.addAuthor(
                    author, role, email, website
                    )
            
            # Create a dialogue window to describe the applet.
            
            self._about_dialog = KAboutApplication(self._about_data)
        
        # Show the dialogue window.
        
        self._about_dialog.show()
    
    # Unimplemented methods (not enabled on class initialisation)
    
    def help(self):
    
        """help(self)
        
        Called when the Help entry is activated from the menu opened
        from the tiny arrow button on the applet's handle.
        
        The appropriate flag must have been set when this class was
        instantiated.
        """
        pass
    
    def preferences(self):
    
        """preferences(self)
        
        Called when the Preferences entry is activated from the menu opened
        from the tiny arrow button on the applet's handle.
        
        The appropriate flag must have been set when this class was
        instantiated.
        """
        pass
    
    def reportBug(self):
    
        """reportBug(self)
        
        Called when the "Report Bug" entry is activated from the menu opened
        from the tiny arrow button on the applet's handle.
        
        The appropriate flag must have been set when this class was
        instantiated.
        """
        pass
    
    # Reimplemented events
    
    def paintEvent(self, event):
    
        """paintEvent(self, event)
        
        Redraw the contents of the applet's widget using the information
        provided in the QPaintEvent object supplied.
        """
        
        # Don't bother to redraw if the widget is not visible.
        
        if not self.isVisible(): return
        
        # Create a painter object.
        
        painter = QPainter(self)
        
        # Draw the pixmap.
        
        area = self.rect()
        
        # Paranoia: if the image was installed then render it, otherwise
        # fill in the applet's panel space in a solid colour.
        
        if hasattr(self, "image"):
            painter.drawImage(area, self.image)
        else:
            painter.fillRect(area, self.brush)
    
    def mousePressEvent(self, event):
    
        """mousePressEvent(self, event)
        
        Handle mouse clicks on the applet's widget described by the
        QMouseEvent object supplied.
        """
        
        # Performs actions depending on the mouse button pressed.
        # Perhaps the buttons for this could be made configurable.
        
        if event.button() == QMouseEvent.LeftButton or \
            event.button() == QMouseEvent.RightButton:
        
            # Show a popup menu.
            self.show_popup()
    
    def read_document(self, this_element):
    
        """[Specifically for this applet]
        
        items = read_document(self, this_element)
        
        Read the relevant items found beneath a DOM element.
        """
        
        items = []
        
        # Find the first child then iterate through its siblings.
        
        child = this_element.firstChild()
        
        while not child.isNull():
        
            element = child.toElement()
            
            if str(element.tagName()) == "folder":
            
                # Read folder titles and examine their contents.
                title = self.read_title(element)
                items.append( (title, self.read_document(element)) )
            
            elif str(element.tagName()) == "bookmark":
            
                # Read bookmark titles and record their destinations.
                title = self.read_title(element)
                href = unicode(element.attribute("href"))
                items.append( (title, href) )
            
            child = child.nextSibling()
        
        # Return the items found to the caller.
        
        return items
    
    def read_title(self, element):
    
        """[Specifically for this applet]
        
        title = read_title(self, element)
        
        Read the value of any title elements found beneath the element
        specified.
        """
        
        first_child = element.firstChild()
        
        if first_child.isNull(): return u""
        
        element = first_child.toElement()
        
        if str(element.tagName()) != "title":
            return u""
        
        return unicode(element.text()).strip()
    
    def load_bookmarks(self):
    
        """[Specifically for this applet]
        
        document = load_bookmarks(self)
        
        Load the bookmarks from the user's bookmarks.xml file from
        $HOME/.kde/share/apps/konqueror/bookmarks.xml and tries to
        create a QDomDocument object to return to the caller. If it
        fails then it returns None instead.
        """
        
        # Find the user's home directory.
        home = os.getenv("HOME")
        
        # Find the bookmarks.xml file beneath the user's home directory.
        path = os.path.join(
            home, ".kde", "share", "apps", "konqueror",
            "bookmarks" + os.extsep + "xml"
            )
        
        try:
            # Open the file using QFile rather than the open function.
            
            f = QFile(path)
            if f == 0: raise IOError, "Not found"
            
            f.open(IO_ReadOnly)
            if f == 0: raise IOError, "Cannot open for reading"
            
            # Create a QDomDocument to contain the bookmark descriptions.
            document = qtxml.QDomDocument()
            
            # Read the file into the document, converting the XML into a
            # DOM representation.
            result, errorMsg, errorLine, errorColumn = document.setContent(f)
            
            if result == 0: raise IOError, "Cannot read document."
        
        except IOError:
        
            # Cannot open the file.
            return None
        
        # Return the document to the caller.
        
        return document
        
    def show_popup(self):
    
        """[Specifically for this applet]
        
        show_popup(self)
        
        Typically called by an event handling method (see mousePressEvent
        above), this method obtains the user's bookmarks file using
        the load_bookmarks method and opens a popup menu containing
        entries for each bookmark with the contents of folders in submenus.
        """
        
        document = self.load_bookmarks()
        
        if document is None:
        
            # Open an error dialog here then return.
            
            KMessageBox.sorry(
                None, i18n("Could not find the bookmarks file.")
                )
            return
        
        items = self.read_document(document.documentElement())
        popup = KPopupMenu(self)
        popup.insertTitle(i18n("Bookmarks"))
        
        # Record the IDs of the menu items as they are created.
        self.ids = {}
        self.popup_items(popup, items)
        result = popup.exec_loop(QCursor.pos())
        
        try:
            url = self.ids[result]
            
            if url is not None:
                self.browser = KParts.BrowserRun(KURL(url), KParts.URLArgs(), None, None, True, False)
        
        except KeyError:
            pass
    
    def popup_items(self, menu, items):
    
        """[Specifically for this applet]
        
        popup_items(self, menu, items)
        
        Examine the contents of the list structure containing details
        of the bookmarks and folders, adding entries to menus for the
        bookmarks and submenus for the folders.
        """
        
        for title, details in items:
        
            if type(details) == type([]):
            
                # Folder
                
                new_menu = KPopupMenu(self)
                new_menu.insertTitle(title)
                self.ids[menu.insertItem(title, new_menu)] = None
                
                self.popup_items(new_menu, details)
            
            else:
            
                # Bookmark
                
                # Impose arbitrary limit on the length of the menu entry.
                
                if len(title) > 32:
                    title = title[:32] + u"..."
                
                self.ids[menu.insertItem(title)] = details
