/*
 * panel_launcher.cpp
 * 
 * Launch panel applets written in Python using an embedded Python interpreter.
 * 
 * Copyright (C) 2003-2006 Jim Bublitz, David Boddie <david@boddie.org.uk>
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include <Python.h>
#include <qobject.h>
#include <kpanelapplet.h>
#include <klibloader.h>
#include "panel_launcher.h"

#define debug 0

KPanelApplet* report_error(char *msg, bool report_python = true)
{
    #if debug == 1
    printf ("error: %s\n", msg);
    #endif

    if (report_python) PyErr_Print();

    return NULL;
}

PyObject* call_function(PyObject *callable, PyObject *args)
{
    PyObject *result = NULL, *pArgs;

    if (callable == NULL)
    {
        return NULL;
    }

    if (PyCallable_Check(callable))
    {
        if (args == NULL)
            pArgs = PyTuple_New(0);
        else
            pArgs = args;

        result = PyObject_CallObject(callable, pArgs);

        /* If the arguments were created in this function then decrease
           their reference count. */
        if (args == NULL)
            Py_XDECREF(pArgs);
            /* pDict and pFunc are borrowed and must not be Py_DECREF-ed */

        if (result == NULL)
        {
            #if debug == 1
            PyErr_Print();
            #else
            PyErr_Clear();
            #endif
        }
    }

    return result;
}

// Declare a Python object to represent the object returned by the
// Python module.
static KPanelApplet *applet = NULL;
static PyObject *pyFactory = NULL;

extern "C"
{
    KPanelApplet* CPP_FACTORY( QWidget *parent, const QString& configFile )
    {
        if (!Py_IsInitialized())
        {
            KLibLoader::self()->globalLibrary(PYTHON_LIB_PATH);
            Py_Initialize();
        }

        // Locate the path to the script and append it to sys.path list.

        PyRun_SimpleString("import sys\n");
        PyRun_SimpleString("sys.path.append('"MODULE_DIR"')\n");

        PyObject *pyModule = PyImport_ImportModule(MODULE_NAME);
        /* pyModule is a new reference */

        if (!pyModule) return report_error ("*** failed to import module\n");

        PyObject *pyDict = PyModule_GetDict(pyModule);
        /* pDict is a borrowed reference */

        Py_XDECREF(pyModule);

        // Get the Python module's factory function from the header file at
        // compile time.
        if (!pyFactory)
            pyFactory = PyDict_GetItemString(pyDict, FACTORY);

        /* pyFactory is a borrowed reference */
        if (!pyFactory) return report_error("*** failed to find factory\n");

        // Set up the args for the Python factory function call.
        PyObject *args = NULL;
        PyObject *pArg1, *pArg2, *pArg3;
        args = PyTuple_New(3);
        
        pArg1 = PyLong_FromVoidPtr((void*)parent);
        pArg2 = PyString_FromString(configFile.ascii());
        pArg3 = PyString_FromString(MODULE_NAME);
        
        PyTuple_SetItem(args, 0, pArg1);
        PyTuple_SetItem(args, 1, pArg2);
        PyTuple_SetItem(args, 2, pArg3);
        
        // Instantiate the class.
        PyObject *pyAppletPtr = call_function (pyFactory, args);
        /* pyAppletPtr should be a new reference to a long object. */

        Py_XDECREF(args);

        if (pyAppletPtr)
        {
            // Convert the PyObject to a C++ object. *
            applet = (KPanelApplet*)PyLong_AsVoidPtr(pyAppletPtr);
        }

        Py_XDECREF(pyAppletPtr);

        return applet;
    }
}
